@@ -28,29 +28,30 @@ our $_CURRENT_CLOSE_DELIMITER;
 
 sub compile {
     my ($class, $ast) = @_;
-    die "Invalid AST: empty AST" unless @$ast;
+    die "Invalid AST: empty AST" unless @$ast; # uncoverable branch true
 
-    my $first_delimiter_syntax = $ast->[0];
+    my @ast = @$ast;
+    my $first_delimiter_syntax = shift @ast;
     my ($type, $open_delimiter, $close_delimiter) = @$first_delimiter_syntax;
-    if ($type != SYNTAX_DELIMITER) {
-        croak "Invalid AST: Delimiter should be first syntax";
+    if ($type != SYNTAX_DELIMITER) { # uncoverable branch true
+        croak "Invalid AST: Delimiter should be first syntax"; # uncoverable statement
     }
 
-    $ast = do {
+    @ast = do {
         local $_DEFAULT_OPEN_DELIMITER = $open_delimiter;
         local $_DEFAULT_CLOSE_DELIMITER = $close_delimiter;
-        _optimize([@{$ast}[1..$#{$ast}]], 0);
+        _optimize(\@ast, 0);
     };
 
     # Optimize: empty
-    return sub { '' } if @$ast == 0;
+    return sub { '' } if @ast == 0;
 
     # Optimize: raw text only
-    if (@$ast == 1 && $ast->[0]->[0] == SYNTAX_RAW_TEXT) {
-        my (undef, $text) = @{ $ast->[0] };
-        if ($text =~ /[\r\n](?!\z)/mano) {
+    if (@ast == 1 && $ast[0][0] == SYNTAX_RAW_TEXT) {
+        my (undef, $text) = @{ $ast[0] };
+        if (!@CONTEXT_HINT && $text =~ /[\r\n](?!\z)/mano) {
             return sub {
-                defined $_PADDING ? $text =~ s/(\r\n?|\n)(?!\z)/${1}${_PADDING}/mgaor : $text
+                defined $_PADDING ? $text =~ s/(\r\n?|\n)(?!\z)/${1}${_PADDING}/mgar : $text
             };
         }
         return sub { $text };
@@ -62,9 +63,9 @@ sub compile {
         local $_DEFAULT_CLOSE_DELIMITER = $close_delimiter;
         local $_CURRENT_OPEN_DELIMITER = $open_delimiter;
         local $_CURRENT_CLOSE_DELIMITER = $close_delimiter;
-        _compile($ast, 4);
+        _compile(\@ast, 4);
     };
-    die "Invalid AST: $@" if "$@";
+    die "Invalid AST: $@" if "$@"; # uncoverable branch true
 
     # wrap to define function global variables
     $code = <<__CODE__;
@@ -74,11 +75,10 @@ do {
 $code
 };
 __CODE__
-
-    warn $code if DEBUG;
+    warn $code if DEBUG; # uncoverable branch true
 
     my $f = eval $code;
-    die $@ if $@;
+    die $@ if $@; # uncoverable branch true
     return $f;
 }
 
@@ -102,7 +102,6 @@ sub _compile {
     $code .= (' ' x $indent)."sub {\n";
     $code .= (' ' x $indent)."    local \@_CTX = \@_;\n";
     $code .= (' ' x $indent)."    local (\$_OPEN_DELIMITER, \$_CLOSE_DELIMITER) = ($default_open_delimiter_perl, $default_close_delimiter_perl);\n";
-    $code .= (' ' x $indent)."    my (\$_current_open_delimiter, \$_current_close_delimiter) = ($current_open_delimiter_perl, $current_close_delimiter_perl);\n";
     $code .= (' ' x $indent)."    local \$Text::MustacheTemplate::Evaluator::LAMBDA_RENDERER = \\&_render_template_in_context;\n" if $indent == 4;
     $code .= "\n";
     $code .= (' ' x $indent)."    my \$_result = $initial_text_perl;\n";
@@ -115,10 +114,9 @@ sub _compile {
 sub _optimize {
     my ($ast, $depth) = @_;
 
-    my @ast = @$ast;
     my $raw_text_syntax;
     my @optimized_ast;
-    while (my $syntax = shift @ast) {
+    for my $syntax (@$ast) {
         if ($syntax->[0] == SYNTAX_RAW_TEXT) {
             if ($raw_text_syntax) {
                 $raw_text_syntax->[1] .= $syntax->[1];
@@ -137,16 +135,16 @@ sub _optimize {
                     local our $_OPEN_DELIMITER = $_DEFAULT_OPEN_DELIMITER;
                     local our $_CLOSE_DELIMITER = $_DEFAULT_CLOSE_DELIMITER;
                     local our @_CTX = @CONTEXT_HINT;
-                    local $Text::MustacheTemplate::Evaluator::LAMBDA_RENDERER = \&_render_template_in_context;
+                    local $Text::MustacheTemplate::Evaluator::LAMBDA_RENDERER = \&_render_template_in_context if $Text::MustacheTemplate::LAMBDA_TEMPLATE_RENDERING;
 
-                    my $value = retrieve_variable(\@_CTX, split /\./ano, $name);
+                    my $value = $name eq '.' ? $_CTX[-1] : retrieve_variable(\@_CTX, split /\./ano, $name);
                     next unless $value;
-                    if ($type == VARIABLE_HTML_ESCAPE) {
+                    if ($type == VARIABLE_HTML_ESCAPE) { # uncoverable branch false count:2
                         $value = escape_html($value);
                     } elsif ($type == VARIABLE_RAW) {
                         # nothing to do
                     } else {
-                        die "Unknown variable type: $type";
+                        die "Unknown variable type: $type"; # uncoverable statement
                     }
 
                     if ($raw_text_syntax) {
@@ -161,14 +159,14 @@ sub _optimize {
                         local our $_OPEN_DELIMITER = $_DEFAULT_OPEN_DELIMITER;
                         local our $_CLOSE_DELIMITER = $_DEFAULT_CLOSE_DELIMITER;
                         local our @_CTX = @CONTEXT_HINT;
-                        local $Text::MustacheTemplate::Evaluator::LAMBDA_RENDERER = \&_render_template_in_context;
-                        next unless evaluate_section_variable(\@_CTX, split /\./ano, $name);
+                        local $Text::MustacheTemplate::Evaluator::LAMBDA_RENDERER = \&_render_template_in_context if $Text::MustacheTemplate::LAMBDA_TEMPLATE_RENDERING;
+                        next unless $name eq '.' ? evaluate_section($_CTX[-1]) : evaluate_section_variable(\@_CTX, split /\./ano, $name);
                     } elsif ($type == BOX_INVERTED_SECTION) {
                         local our $_OPEN_DELIMITER = $_DEFAULT_OPEN_DELIMITER;
                         local our $_CLOSE_DELIMITER = $_DEFAULT_CLOSE_DELIMITER;
                         local our @_CTX = @CONTEXT_HINT;
-                        local $Text::MustacheTemplate::Evaluator::LAMBDA_RENDERER = \&_render_template_in_context;
-                        next if evaluate_section_variable(\@_CTX, split /\./ano, $name);
+                        local $Text::MustacheTemplate::Evaluator::LAMBDA_RENDERER = \&_render_template_in_context if $Text::MustacheTemplate::LAMBDA_TEMPLATE_RENDERING;
+                        next if $name eq '.' ? evaluate_section($_CTX[-1]) : evaluate_section_variable(\@_CTX, split /\./ano, $name);
                     }
                 }
             }
@@ -178,9 +176,9 @@ sub _optimize {
                 $raw_text_syntax = undef;
             }
             if ($syntax->[0] == SYNTAX_BOX) {
-                my $children = _optimize($syntax->[-1], $depth+1);
+                my @children = _optimize($syntax->[-1], $depth+1);
                 $syntax = [@$syntax]; # shallow copy
-                $syntax->[-1] = $children;
+                $syntax->[-1] = \@children;
             }
             push @optimized_ast => $syntax;
         }
@@ -188,24 +186,24 @@ sub _optimize {
     if ($raw_text_syntax) {
         push @optimized_ast => $raw_text_syntax;
     }
-    return \@optimized_ast;
+    return @optimized_ast;
 }
 
 sub _compile_body {
     my ($ast, $indent, $result) = @_;
-    my @ast = @$ast;
 
     my $code = '';
-    while (my $syntax = shift @ast) {
+    for my $i (keys @$ast) {
+        my $syntax = $ast->[$i];
         my ($type) = @$syntax;
-        if ($type == SYNTAX_RAW_TEXT) {
+        if ($type == SYNTAX_RAW_TEXT) { # uncoverable branch false count:5
             my (undef, $text) = @$syntax;
             next if $result eq DISCARD_RESULT;
-            if (@ast ? $text =~ /[\r\n]/mano : $text =~ /[\r\n](?!\z)/mano) {
+            if ($i == $#{$ast} ? $text =~ /[\r\n](?!\z)/mano : $text =~ /[\r\n]/mano) {
                 my $regex = '(\r\n?|\n)';
-                $regex .= '(?!\z)' unless @ast;
+                $regex .= '(?!\z)' if $i == $#{$ast};
                 $code .= (' ' x $indent).'$_tmp = '.B::perlstring($text).";\n";
-                $code .= (' ' x $indent)."\$_tmp =~ s/$regex/\${1}\${_PADDING}/mago if defined \$_PADDING;\n";
+                $code .= (' ' x $indent)."\$_tmp =~ s/$regex/\${1}\${_PADDING}/mag if defined \$_PADDING;\n";
                 $code .= (' ' x $indent)."$result .= \$_tmp;\n";
             } else {
                 $code .= (' ' x $indent).$result.' .= '.B::perlstring($text).";\n";
@@ -213,7 +211,6 @@ sub _compile_body {
         } elsif ($type == SYNTAX_DELIMITER) {
             my (undef, $open_delimiter, $close_delimiter) = @$syntax;
             ($_CURRENT_OPEN_DELIMITER, $_CURRENT_CLOSE_DELIMITER) = ($open_delimiter, $close_delimiter);
-            $code .= (' ' x $indent)."(\$_current_open_delimiter, \$_current_close_delimiter) = (".B::perlstring($open_delimiter).", ".B::perlstring($close_delimiter).");\n";
         } elsif ($type == SYNTAX_VARIABLE) {
             $code .= _compile_variable($syntax, $indent, $result);
         } elsif ($type == SYNTAX_BOX) {
@@ -233,7 +230,7 @@ sub _compile_body {
             $code .= (' ' x $indent)."    \$Text::MustacheTemplate::REFERENCES{\$_name}->(\@_CTX);\n";
             $code .= (' ' x $indent)."} if exists \$Text::MustacheTemplate::REFERENCES{\$_name};\n";
         } else {
-            die "Unknown syntax: $type";
+            die "Unknown syntax: $type"; # uncoverable statement
         }
     }
     return $code;
@@ -243,14 +240,14 @@ sub _compile_variable {
     my ($syntax, $indent, $result) = @_;
 
     my (undef, $type, $name) = @$syntax;
-    if ($type == VARIABLE_HTML_ESCAPE) {
+    if ($type == VARIABLE_HTML_ESCAPE) { # uncoverable branch false count:2
         my $retriever = $name eq '.' ? '$_CTX[-1]' : 'retrieve_variable(\@_CTX, '.(join ', ', map B::perlstring($_), split /\./ano, $name).')';
         return (' ' x $indent)."$result .= escape_html($retriever // '');\n";
     } elsif ($type == VARIABLE_RAW) {
         my $retriever = $name eq '.' ? '$_CTX[-1]' : 'retrieve_variable(\@_CTX, '.(join ', ', map B::perlstring($_), split /\./ano, $name).')';
         return (' ' x $indent)."$result .= $retriever // '';\n";
     } else {
-        die "Unknown variable: $type";
+        die "Unknown variable: $type"; # uncoverable statement
     }
 }
 
@@ -258,20 +255,11 @@ sub _compile_box {
     my ($syntax, $indent, $result) = @_;
 
     my (undef, $type) = @$syntax;
-    if ($type == BOX_SECTION) {
+    if ($type == BOX_SECTION) { # uncoverable branch false count:4
         my (undef, undef, $name, $inner_template, $children) = @$syntax;
         my $no_lambda = @CONTEXT_HINT && !$Text::MustacheTemplate::LAMBDA_TEMPLATE_RENDERING;
 
-        my ($open_delimiter, $close_delimiter) = ($_CURRENT_OPEN_DELIMITER, $_CURRENT_CLOSE_DELIMITER);
         my $inner_code = _compile_body($children, $no_lambda ? $indent+4 : $indent+8, $result);
-
-        # reset delimiter before loop
-        my $delimiter_reset;
-        if ($_CURRENT_OPEN_DELIMITER ne $open_delimiter || $_CURRENT_CLOSE_DELIMITER ne $close_delimiter) {
-            $delimiter_reset = "(\$_current_open_delimiter, \$_current_close_delimiter) = (".B::perlstring($_CURRENT_OPEN_DELIMITER).", ".B::perlstring($_CURRENT_CLOSE_DELIMITER).");\n";            
-        }
-
-        $inner_template = B::perlstring($inner_template);
         my $evaluator = $name eq '.'
             ? 'evaluate_section($_CTX[-1])'
             : 'evaluate_section_variable(\@_CTX, '.(join ', ', map B::perlstring($_), split /\./ano, $name).')';
@@ -280,24 +268,25 @@ sub _compile_box {
             my $code = (' ' x $indent)."push \@_CTX => {};\n";
             $code .= (' ' x $indent)."for my \$ctx ($evaluator) {\n";
             $code .= (' ' x $indent)."    \$_CTX[-1] = \$ctx;\n";
-            $code .= (' ' x $indent)."    $delimiter_reset" if defined $delimiter_reset;
-            $code .=                          $inner_code;
+            $code .=                      $inner_code;
             $code .= (' ' x $indent)."}\n";
             $code .= (' ' x $indent)."pop \@_CTX;\n";
+            return $code;
         }
 
+        my ($open_delimiter, $close_delimiter) = map B::perlstring($_), ($_CURRENT_OPEN_DELIMITER, $_CURRENT_CLOSE_DELIMITER);
+        $inner_template = B::perlstring($inner_template);
         my $code = (' ' x $indent)."\@_section = $evaluator;\n";
         $code .= (' ' x $indent)."if (\$Text::MustacheTemplate::LAMBDA_TEMPLATE_RENDERING && \@_section == 1 && ref \$_section[0] eq 'CODE') {\n";
         $code .= (' ' x $indent)."    my \$code = \$_section[0];\n";
         $code .= (' ' x $indent)."    \$_tmp = \$code->($inner_template);\n";
-        $code .= (' ' x $indent)."    local (\$_OPEN_DELIMITER, \$_CLOSE_DELIMITER) = (\$_current_open_delimiter, \$_current_close_delimiter);\n";
+        $code .= (' ' x $indent)."    local (\$_OPEN_DELIMITER, \$_CLOSE_DELIMITER) = ($open_delimiter, $close_delimiter);\n";
         $code .= (' ' x $indent)."    $result .= _render_template_in_context(\$_tmp);\n";
         $code .= (' ' x $indent)."} else {\n";
         $code .= (' ' x $indent)."    my \@section = \@_section;\n"; # copy to avoid rewrite same varialbe in recurse
         $code .= (' ' x $indent)."    push \@_CTX => {};\n";
         $code .= (' ' x $indent)."    for my \$ctx (\@section) {\n";
         $code .= (' ' x $indent)."        \$_CTX[-1] = \$ctx;\n";
-        $code .= (' ' x $indent)."        $delimiter_reset" if defined $delimiter_reset;
         $code .=                          $inner_code;
         $code .= (' ' x $indent)."    }\n";
         $code .= (' ' x $indent)."    pop \@_CTX;\n";
@@ -330,11 +319,6 @@ sub _compile_box {
         my $code = (' ' x $indent)."unless (exists \$_BLOCKS{$name}) {\n";
         $code .= (' ' x $indent)."    \$_BLOCKS{$name} = $sub_code";
         $code .= (' ' x $indent)."}\n";
-
-        # apply inner delimiter change into current scope
-        if ($_CURRENT_OPEN_DELIMITER ne $open_delimiter || $_CURRENT_CLOSE_DELIMITER ne $close_delimiter) {
-            $code .= (' ' x $indent)."(\$_current_open_delimiter, \$_current_close_delimiter) = (".B::perlstring($_CURRENT_OPEN_DELIMITER).", ".B::perlstring($_CURRENT_CLOSE_DELIMITER).");\n";            
-        }
         return $code;
     } elsif ($type == BOX_PARENT) {
         local $_PARENT = $syntax;
@@ -354,14 +338,14 @@ sub _compile_box {
         $code .= (' ' x $indent)."}\n";
         return $code;
     } else {
-        die "Unknown box: $type";
+        die "Unknown box: $type"; # uncoverable statement
     }
 }
 
 sub _render_template_in_context {
     my $source = shift;
     our ($_OPEN_DELIMITER, $_CLOSE_DELIMITER);
-    if ($source !~ /(?:\Q$_OPEN_DELIMITER\E|\Q$_CLOSE_DELIMITER\E)/mano) {
+    if ($source !~ /(?:\Q$_OPEN_DELIMITER\E|\Q$_CLOSE_DELIMITER\E)/man) {
         return $source;
     }
 
